"""
PyPorCC

========

PyPorCC module. Classifier for harbour porpoises clicks

Detection of NBHF clicks of harbour porpoises using SoundTrap.

These NBHF signals have comparable durations around 100
μsec, high directionality, centre frequencies around 130 kHz, and
source2 levels generally well below 200 dB re 1 μPa (Au 1993; Madsen
et al. 2005; Villadsgaard et al. 2007; Kyhn et al. 2009). 

Clicks were only selected when the signal-to-noise ratio (SNR), given by the ratio
between the rms power of the signal of interest and the ambient
noise level measured in the same frequency band as the signal, was
at least 10 dB. Also, the first and last clicks in a click train were
identified if possible, to include as much of the entire click train as
possible in the analysis.

Prior to analysis, the sound files were filtered with a high-pass
digital Butterworth filter (4th order, 100 kHz –3dB cut off frequency). 

Click repetition rates were either measured using an automated click
detection algorithm made in Matlab or, for long ICI click trains, by
computing the pulse repetition spectrum (Watkins 1967). 


The duration of individual clicks ranges from
50 ls to 175 ls and the half-power (3 dB) bandwidth is
around 15 kHz (Kyhn et al., 2010). Clicks are emitted in
series, often referred to as “trains.” A click train is loosely
defined as “any series of clicks separated by gradually or
cyclically changing inter-click interval suggesting a unit during an echolocation event or a communication signal”

"""
from pyporcc.click_detector import ClickDetector, ClickDetectorSoundTrapHF, Filter
from pyporcc.click_converter import Click, ClickConverter
from pyporcc.porpoise_classifier import PorpoiseClassifier
from pyporcc.porcc import PorCCModel, PorCC



#!/usr/bin/python
"""
Module : utils.py
Authors : Clea Parcerisas
Institution : VLIZ (Vlaams Instituut voor de Zee)
Last Accessed : 9/23/2020
"""

__author__ = "Clea Parcerisas"
__version__ = "0.1"
__credits__ = "Clea Parcerisas"
__email__ = "clea.parcerisas@vliz.be"
__status__ = "Development"

import numba as nb
import numpy as np

REF = 1.0


@nb.jit
def aic_score(y, y_prob, n_features):
    """
    Return the AIC score
    """
    llk = np.sum(y * np.log(y_prob[:, 1]) + (1 - y) * np.log(y_prob[:, 0]))
    aic = 2 * n_features - 2 * llk
    return aic


@nb.jit
def to_upa(wave, sensitivity, preamp_gain, Vpp):
    """
    Return the wave in db
    Parameters
    ----------
    wave
    sensitivity
    preamp_gain
    Vpp

    Returns
    -------
    The same wave converted to upa
    """
    mv = 10 ** (sensitivity / 20.0) * REF
    ma = 10 ** (preamp_gain / 20.0) * REF
    gain_upa = (Vpp / 2.0) / (mv * ma)
    return wave * gain_upa


@nb.jit
def amplitude_db(clip):
    """
    Return the amplitude of the clip
    Parameters
    ----------
    clip : np.array
        Signal
    Returns
    -------
    The maximum amplitude of the clip in db
    """
    return to_db(np.max(np.abs(clip)))


@nb.jit
def to_db(wave):
    """
    Convert the wave to db
    Parameters
    ----------
    wave : np.array
        wave to compute

    Returns
    -------
    wave in db
    """
    return 10 * np.log10(wave ** 2)


def constrain(x, lower, upper):
    """
    MATLAB constrain
    Constrain between upper and lower limits, and do not ignore NaN
    """
    if isinstance(x, float):
        if x < lower:
            x = lower
        elif x > upper:
            x = upper
    else:
        x[np.where(x < lower)[0]] = lower
        x[np.where(x > upper)[0]] = upper

    return x



#!/usr/bin/python
"""
Module : click_converter.py
Authors : Clea Parcerisas
Institution : VLIZ (Vlaams Instituut voor de Zee)
Last Accessed : 9/23/2020
"""

__author__ = "Clea Parcerisas"
__version__ = "0.1"
__credits__ = "Clea Parcerisas"
__email__ = "clea.parcerisas@vliz.be"
__status__ = "Development"

import sys
from importlib import resources

import matplotlib.pyplot as plt
import numba as nb
import numpy as np
import soundfile as sf
from scipy import interpolate
from scipy import signal as sig


class ClickConverter:
    def __init__(self, fs, click_model_path=None, click_vars=None):
        """
        Init the click converter

        Parameters
        ----------
        click_model_path : string or Path
            File where the click model is stored
        click_vars : list of strings
            List of the output parameters to compute for each click
        """
        if click_model_path is None:
            with resources.path('pyporcc.data', 'standard_click.wav') as click_model_path:
                print('Setting the click model path to default...')
                self.click_model_path = click_model_path
        self.click_model, self.fs_model = sf.read(click_model_path)
        self.fs = fs

        if click_vars is None:
            self.click_vars = ['Q', 'duration', 'ratio', 'XC', 'CF', 'BW']
        else:
            self.click_vars = click_vars

    def __setattr__(self, key, value):
        if key == 'fs':
            if self.fs_model != value:
                click_model, fs_model = sf.read(self.click_model_path)
                print('This click is not recorded at the same frequency than the classified data! '
                      'Resampling to %s S/s' % value)
                new_samples = int(np.ceil(click_model.size / fs_model * value))
                self.click_model = sig.resample(click_model, new_samples)
                self.fs_model = value
        self.__dict__[key] = value

    def clicks_df(self, df, nfft=512, save_path=None):
        """
        Find all the clicks and return them as a df with all the clicks params

        Parameters
        ----------
        df : DataFrame
            It needs to have at least datetime and wave
        nfft : int
            Length of FFT
        save_path : string or Path
            Path to the desired save file. If None, it is not saved
        """
        # Calculate all the independent variables and append them to the df
        params_matrix = np.zeros((len(df), len(self.click_vars)))

        for idx in df.index:
            row = df.iloc[idx]
            params_matrix[idx, :] = self.click_params(row['wave'], nfft=nfft)

        df[self.click_vars] = params_matrix

        # Keep the metadata
        df.fs = df.fs

        if save_path is not None:
            extension = save_path.split('.')[-1]
            if extension == 'json':
                df.to_json(save_path)
            elif extension == 'pkl':
                df.to_pickle(save_path)
            elif extension == 'csv':
                df.drop(['wave'], axis=1).to_csv(save_path)
            elif extension == 'hdf5':
                df.to_hdf(save_path, mode='w', key='clips')
            else:
                raise Exception('The extension %s is unkown' % extension)

        return df

    def row2click(self, row):
        """
        Convert the row to an object Click
        Parameters
        ----------
        row : Pandas DataFrame row

        Returns
        -------
        Click object
        """
        signal = row['wave']
        if 'datetime' in row.axes[0]:
            timestamp = row.datetime
        else:
            timestamp = row.name
        click = Click(signal, self.fs, timestamp, click_model_path=self.click_model_path, verbose=False)
        return click

    def add_params_to_row(self, row, nfft=512):
        row.loc[self.click_vars] = self.click_params(row['wave'], nfft)
        return row

    def click_params(self, sound_block, nfft=512):
        q, duration, ratio, xc, cf, bw, _, _ = click_params(sound_block, self.fs, self.click_model, nfft)
        return q, duration, ratio, xc, cf, bw

    @staticmethod
    def test_click_calculation(df_clicks, df_test, col_vars):
        """
        Test the calculation of the click parameters indicated in col_vars obtained with python
        compared to the ones obtained in the paper (on the Test DB)
        """
        # Compare each field
        rel_error = np.abs(df_test[col_vars] - df_clicks[col_vars]) / df_test[col_vars].mean()
        mean_rel_error = rel_error.mean()

        return mean_rel_error


@nb.njit
def zero_pad(sound_block, nfft):
    """
    Return a zero-padded sound block
    Parameters
    ----------
    sound_block: np.array
        Sound clip representing the possible click
    nfft : int
        Desired length
    """
    if len(sound_block) < nfft:
        zero_padded = np.zeros(nfft)
        zero_padded[0:sound_block.size] = sound_block
        sound_block = zero_padded
    return sound_block


@nb.jit(nopython=True)
def fast_click_params(sound_block, fs, psd, freq):
    """
    Return the duration of the 80% of the energy of the sound block
    Parameters
    ----------
    sound_block: np.array
        Sound Block to analyze
    fs : int
        Sampling Frequency
    psd : np.array
        Power Spectral Density
    freq : np.array
        Frequencies of the psd

    Returns
    -------
    Duration in microseconds
    """
    ener = np.cumsum(sound_block ** 2)
    istart_list = np.where(ener <= (ener[-1] * 0.1))[0]  # index of where the 10% is
    iend_list = np.where(ener <= (ener[-1] * 0.9))[0]  # index of where the 90% is
    if len(istart_list) > 0:
        istart = istart_list[-1]
    else:
        istart = 0
    if len(iend_list) > 0:
        iend = iend_list[-1]
    else:
        iend = len(ener)
    # duration in microseconds
    duration = iend - istart
    duration = duration / fs * 1e6

    # Normalize spectrum
    psd = psd / np.max(psd)
    cf = np.sum(freq * (psd ** 2)) / np.sum(psd ** 2)
    pf = freq[psd.argmax()]

    # Calculate RMSBW
    rmsbw = (np.sqrt((np.sum(((freq - cf) ** 2) * (psd ** 2))) / np.sum(psd ** 2))) / 1000.0

    # Parameters according to Mhl & Andersen, 1973
    q = (cf / rmsbw) / 1000.0
    ratio = pf / cf

    return duration, cf, pf, q, ratio


def powerbw(x, fs):
    """
    Calculate -3dB bandwidth imitating MATLAB powerbw
    Parameters
    ----------
    x : np.array
        Signal to process
    fs : int
        Sampling frequency

    Returns
    -------
    -3dB Bandwidth in kHz
    """
    beta = 0
    freq, psd = sig.periodogram(x=x, fs=fs, window=('kaiser', beta), scaling='spectrum',
                                return_onesided=False)

    # Consecutive frequencies of the psd that have more than half of the maximum freq power
    r = 10 * np.log10(0.5)
    half = np.max(psd) * (10 ** (r / 10))
    i_center = psd.argmax()

    i_l = np.where(psd[0:i_center] <= half)[0]
    i_r = np.where(psd[i_center:-1] <= half)[0]

    if len(i_l) == 0:
        f_lo = freq[0]
    else:
        i_l = i_l[-1]
        psd_points = np.log10([max(psd[i_l], sys.float_info.min), max(psd[i_l+1], sys.float_info.min)])
        inter = interpolate.interp1d(psd_points, freq[i_l: i_l + 2])
        f_lo = inter(np.log10(half))

    if len(i_r) == 0:
        f_hi = freq[-1]
    else:
        i_r = i_r[0] + i_center
        psd_points = np.log10([max(psd[i_r - 1], sys.float_info.min), max(psd[i_r], sys.float_info.min)])
        inter = interpolate.interp1d(psd_points, freq[i_r - 1: i_r + 1])
        f_hi = inter(np.log10(half))

    bw = (f_hi - f_lo) / 1000.0

    return bw


def click_params(sound_block, fs, click_model, nfft):
    """
    Return the click parameters necessary to classify the click using PorCC
    Parameters
    ----------
    sound_block
    fs
    click_model
    nfft

    Returns
    -------
    Q, duration, ratio, XC, CF, BW, psd and freq
    """
    # Calculate PSD, freq, centrum-freq (cf), peak-freq (pf) of the sound file
    sound_block_padded = zero_pad(sound_block, nfft)
    freq, psd = sig.periodogram(x=sound_block_padded, fs=fs, nfft=nfft, scaling='spectrum')
    duration, cf, pf, q, ratio = fast_click_params(sound_block, fs, psd, freq)
    bw = powerbw(sound_block, fs)

    # Calculate the correlation with the model
    x_coeff = np.correlate(sound_block, click_model, mode='same')
    xc = np.max(x_coeff)

    return q, duration, ratio, xc, cf, bw, psd, freq


class Click:
    def __init__(self, sound_block, fs, timestamp, click_model_path, nfft=512, verbose=False):
        """
        Start a porpoise click object

        Parameters
        ----------
        sound_block : np.array
            Clip of the click
        fs : int
            Sampling Frequency
        timestamp : datetime
            Timestamp where the click starts
        click_model_path : string or Path
            File where the click model is stored
        nfft : int
            Number of frequency bins to use for the spectrogram calculation
        verbose : bool
            Set to True if plots are desired
        """
        self.nfft = nfft
        self.fs = fs
        self.sound_block = sound_block
        self.timestamp = timestamp
        self.click_model, fs_model = sf.read(click_model_path)
        if fs_model != fs:
            if verbose:
                print('This click is not recorded at the same frequency than the classified data! '
                      'Resampling to %s S/s' % self.fs)
            new_samples = int(np.ceil(self.click_model.size / fs_model * self.fs))
            self.click_model = sig.resample(self.click_model, new_samples)
        else:
            self.fs = fs

        self.q, self.duration, self.ratio, self.xc, self.cf, self.bw, psd, self.freq = click_params(sound_block,
                                                                                                    self.fs,
                                                                                                    self.click_model,
                                                                                                    nfft)
        if verbose:
            fig, ax = plt.subplots(2, 1)
            ax[0].plot(self.sound_block)
            ax[1].plot(self.freq, psd)
            plt.show()
            plt.close()

    def __getattribute__(self, name):
        """
        Avoid confusion with capital and low letters
        """
        return super().__getattribute__(name.lower())



#!/usr/bin/python
"""
Module : porpoise_classifier.py
Authors : Clea Parcerisas
Institution : VLIZ (Vlaams Instituut voor de Zee)
Last Accessed : 9/23/2020
"""

__author__ = "Clea Parcerisas"
__version__ = "0.1"
__credits__ = "Clea Parcerisas"
__email__ = "clea.parcerisas@vliz.be"
__status__ = "Development"

import pickle

import matplotlib.pyplot as plt
import pandas as pd
import scipy.stats as stats
from sklearn import metrics, linear_model, ensemble, svm, neighbors, pipeline
from sklearn import preprocessing, feature_selection, model_selection, utils

from pyporcc import utils


class PorpoiseClassifier:
    def __init__(self, train_data=None, test_data=None, ind_vars=None, dep_var=None, class_column='model'):
        """
        Find the click model

        Parameters
        ----------
        train_data : DataFrame
            Data to train.
        test_data : DataFrame
            Data to test
        ind_vars : list or np.array
            List of the independent variables to consider (and names of the columns)
        dep_var : string
            Name of the dependent variable (name of the column)
        """
        if ind_vars is None:
            self.ind_vars = ['Q', 'duration', 'ratio', 'XC', 'CF', 'BW']
        else:
            self.ind_vars = ind_vars
        if dep_var is None:
            self.dep_var = 'class'
        else:
            self.dep_var = dep_var
        self.train_data = train_data
        self.test_data = test_data
        self.models = {}
        self.class_column = class_column

    def split_data(self, data_df, train_size=0.4):
        """
        Split the data in train and test and save it to the appropriate attributes of the class

        Parameters
        ----------
        data_df : DataFrame
            DataFrame with all the data to be divided
        train_size : float
            Float between 0 and 1 representing the percentage of training data
        """
        self.train_data, self.test_data = model_selection.train_test_split(data_df, train_size=train_size)
        return self.train_data, self.test_data

    def get_best_model(self, model_name, binary=False, standarize=False, feature_sel=False):
        """
        Train all the classifiers.

        Parameters
        ----------
        model_name : string
            Name of the model to test.
            The implemented ones are
            `svc`: Support Vector Machines
            `lsvc`: Linear Support Vector Machines
            `RandomForest`: Random Forest
            `knn`: K-Nearest Neighbor
        binary : bool
            If true the variables are converted to 0 / 1 (no porpoise / porpoise) instead of noise/lq/hq clicks
        standarize : bool
            Set to True if the variables should be standarized before training
        feature_sel : bool
            Set to True if the best features should be selected instead of all of them

        Returns
        -------
        Dictionary with the name of the model as key and another dictionary as value
        with ind_vars, model, binary as keys and their respective representations in values
        (It also adds it to the property "models" of the class)
        """
        x = self.train_data[self.ind_vars]
        y = self.train_data[self.dep_var]
        if binary:
            # Convert the classes in 0 (no porpoise) or 1 (porpoise)
            y = self.convert2binary(y)

        # If standarize is considered, append it to the pipeline steps
        steps = []
        if standarize:
            # Standarize the data
            scaler = preprocessing.StandardScaler()
            steps.append(('scaler', scaler))

        # Some common parameters
        tol = 1e-3
        gamma = utils.fixes.loguniform(1e-4, 1000)
        c_values = utils.fixes.loguniform(0.1, 1000)
        class_weight = ['balanced', None]

        # Get the model
        if model_name == 'svc':
            # List all the possible parameters that want to be checked
            kernel_list = ['poly', 'rbf']
            degree = stats.randint(1, 4)
            param_distr = {'degree': degree, 'C': c_values, 'gamma': gamma, 'kernel': kernel_list}

            # Classifier with fixed values
            clf = svm.SVC(tol=tol, cache_size=500, probability=True, max_iter=500)

        elif model_name == 'logit':
            # List all the possible parameters that want to be checked
            penalty = ['l1', 'l2', 'elasticnet', 'none']
            param_distr = {'penalty': penalty, 'C': c_values, 'class_weight': class_weight}
            # Classifier with fixed values
            clf = linear_model.LogisticRegression()

        elif model_name == 'forest':
            # List all the possible parameters that want to be checked
            n_estimators = stats.randint(100, 300)
            param_distr = {'n_estimators': n_estimators}
            # Classifier with fixed values
            clf = ensemble.RandomForestClassifier()

        elif model_name == 'knn':
            # List all the possible parameters that want to be checked
            n_neighbors = stats.randint(2, 9)
            algorithm = ['auto', 'ball_tree', 'kd_tree']
            param_distr = {'n_neighbors': n_neighbors, 'algorithm': algorithm}
            # Classifier with fixed values
            clf = neighbors.KNeighborsClassifier()

        else:
            raise Exception('%s is not implemented!' % model_name)

        if feature_sel:
            # selection = feature_selection.RFECV(estimator=svm.LinearSVC(), step=1, scoring='roc_auc')
            selection = feature_selection.SelectFromModel(ensemble.ExtraTreesClassifier(n_estimators=50))
            # selection = feature_selection.SelectFromModel(svm.LinearSVC())
            # Add the feature selection to the steps
            steps.append(('feature_selection', selection))

        # Search for the best parameters
        gm_cv = model_selection.RandomizedSearchCV(estimator=clf, scoring='roc_auc',
                                                   param_distributions=param_distr, n_iter=100)
        steps.append(('classification', gm_cv))

        # Create pipeline and fit
        model = pipeline.Pipeline(steps)
        model.fit(x, y)

        if feature_sel:
            ind_vars = model['feature_selection'].transform(self.test_data[self.ind_vars])
        else:
            ind_vars = self.ind_vars

        print(model['classification'].best_estimator_)
        self.models[model_name] = {'ind_vars': ind_vars, 'model': model, 'binary': binary}

        # Save the model as a pickle file! 
        pickle.dump(model, open('pyporcc/models/%s.pkl' % model_name, 'wb'))
        return self.models[model_name]

    def train_models(self, model_list, binary=False, standarize=False, feature_sel=False):
        """
        Train all the models of the list

        Parameters
        ----------
        model_list : list of strings
            All the models that want to be tested. They have to be the implemented types (see get_best_model)
        binary : bool
            If true the variables are converted to 0 / 1 (no porpoise / porpoise) instead of noise/lq/hq clicks
        standarize : bool
            Set to True if the variables should be standarized before training
        feature_sel : bool
            Set to True if the best features should be selected instead of all of them

        Returns
        -------
        Dictionary with the name of each of the models of the list as keys and another dictionary as value
        with ind_vars, model, binary as keys and their respective representations in values
        (It also adds it to the property "models" of the class)
        """
        for model_name in model_list:
            self.get_best_model(model_name, binary, standarize, feature_sel)
        return self.models

    def test_models(self):
        """
        Test the models

        Returns
        -------
        DataFrame with name, roc_auc, recall and aic as columns
        """
        results = pd.DataFrame(columns=['name', 'roc_auc', 'recall', 'aic'])
        for model_name, model_item in self.models.items():
            ind_vars = model_item['ind_vars']
            model = model_item['model']
            y_test = self.test_data[self.dep_var]
            if model_item['binary']:
                # Convert the classes in 0 (no porpoise) or 1 (porpoise)
                y_test = self.convert2binary(y_test)
            y_pred = model.predict(self.test_data[ind_vars])
            y_prob = model.predict_proba(self.test_data[ind_vars])

            print(metrics.classification_report(y_test, y_pred))

            # Calculate rou_aub, recall and aic
            roc_auc = metrics.roc_auc_score(y_test, y_pred)
            recall = metrics.recall_score(y_test, y_pred)
            aic = utils.aic_score(y_test, y_prob, len(ind_vars))
            results.loc[results.size] = [model_name, roc_auc, recall, aic]

        print(results)
        return results

    def plot_roc_curves(self, porcc_al):
        """
        Plot the roc curves for HQ vs Noise, LQ vs Noise and All vs Noise

        Parameters
        ----------
        porcc_al : object PorCCModel
            Model to compare it to
        """
        fig, ax = plt.subplots(1, 3)
        hq_noise = self.test_data.loc[self.test_data[self.dep_var] != 2]
        lq_noise = self.test_data.loc[self.test_data[self.dep_var] != 1]
        self._plot_roc_curves(porcc_al=porcc_al, df=hq_noise, ax=ax[0])
        ax[0].set_title('HQ vs Noise')
        self._plot_roc_curves(porcc_al=porcc_al, df=lq_noise, ax=ax[1])
        ax[1].set_title('LQ vs Noise')
        self._plot_roc_curves(porcc_al=porcc_al, df=self.test_data, ax=ax[2])
        ax[2].set_title('All vs Noise')

        plt.tight_layout()
        plt.show()
        plt.close()

    def _plot_roc_curves(self, porcc_al, df, ax):
        """
        Plot the roc curves of all the models

        Parameters
        ----------
        porcc_al : object PorCCModel
            Model to compare it to
        df : DataFrame
            DataFrame to consider
        ax : matplotlib ax
            Axis where to plot the ROC curves
        """
        y_test = df[self.dep_var]
        for model_name, model_item in self.models.items():
            ind_vars = model_item['ind_vars']
            model = model_item['model']
            x_test = df[ind_vars]
            if model_item['binary']:
                # Convert the classes in 0 (no porpoise) or 1 (porpoise)
                y_test = self.convert2binary(y_test)
            y_prob = model.predict_proba(x_test)[:, 1]
            fpr, tpr, thresholds = metrics.roc_curve(y_test, y_prob)
            ax.plot(fpr, tpr, label=model_name)

        porcc_al._plot_roc_curves(df, dep_var=self.dep_var, ax0=ax)
        ax.set_xlabel('False alarm rate')
        ax.set_ylabel('Hit rate')
        ax.legend()
        return ax

    @staticmethod
    def convert2binary(y):
        """
        Return the y sample as binary (no porpoise / porpoise)

        Parameters
        ----------
        y : np.array
            Array with 3 classes to be converted to 2 classes (from Noise, LQ and HQ to Porpoise/No Porpoise)
        """
        y = y.replace(3, 0)
        y = y.replace(2, 1)
        return y

    def add_model(self, model_name, model, ind_vars, binary=False):
        """
        Add a model to the models dict

        Parameters
        ----------
        model_name : string
            Name of the new model
        model : object
            Model object already fitted
        ind_vars : list or np.array of strings
            Names of the columns of the independent variables
        binary : bool
            Set to True if the dependent variable should be converted to a binary class
        """
        self.models[model_name] = {'ind_vars': ind_vars, 'model': model, 'binary': binary}

    def save(self, path):
        """
        Save the current models in a file. Should be a pickle extension

        Parameters
        ----------
        path : string or Path
            Where to store the self object
        """
        extension = path.split('.')[-1]
        if extension == 'pkl':
            pickle.dump(self, open(path, 'wb'))
        else:
            raise Exception('The %s extension is unknown!' % extension)

    def classify_click(self, click):
        """
        Classify the click in HQ, LQ, N

        Parameters
        ----------
        click : Click object
            Click to classify

        Returns
        -------
        Class predictions for each of the models calculated
        """
        x = pd.DataFrame(data={'Q': click.Q, 'duration': click.duration, 'ratio': click.ratio, 'XC': click.xc,
                               'CF': click.cf, 'PF': click.pf, 'BW': click.bw})
        porps = self.classify_row(x)
        return porps

    def classify_row(self, x):
        """
        Classify one row according to the params [PF, CF, Q, XC, duration, ratio, BW]

        Parameters
        ----------
        x : DataFrame row
            Needs to have the independent variables needed for all the trained models

        Returns
        -------
        A DataFrame with the class prediction for all the trained models
        """
        for model_info in self.models:
            x[self.class_column] = model_info['model'].predict(x[model_info['ind_vars']])
        return x

    def classify_matrix(self, df):
        """
        Classify according to the params

        Parameters
        ----------
        df : DataFrame
            DataFrame to classify

        Returns
        -------
        A DataFrame with the class prediction for all the trained models
        """
        for model_info in self.models:
            df[self.class_column] = model_info['model'].predict(df[model_info['ind_vars']])
        return df



#!/usr/bin/python
"""
Module : click_detector.py
Authors : Clea Parcerisas
Institution : VLIZ (Vlaams Instituut voor de Zee)
Last Accessed : 9/23/2020
"""

__author__ = "Clea Parcerisas"
__version__ = "0.1"
__credits__ = "Clea Parcerisas"
__email__ = "clea.parcerisas@vliz.be"
__status__ = "Development"

import datetime as dt
import pathlib
import zipfile

import h5py
import matplotlib.pyplot as plt
import numba as nb
import numpy as np
import pandas as pd
import pyhydrophone as pyhy
import soundfile as sf
from scipy import signal as sig
from tqdm import tqdm

from pyporcc import click_converter as cc
from pyporcc import utils


class ClickDetector:
    def __init__(self, hydrophone=None, long_filt=0.00001, long_filt2=0.000001, short_filt=0.1, threshold=10,
                 min_separation=100, max_length=1024, min_length=90, pre_samples=40, post_samples=40, fs=576000,
                 prefilter=None, dfilter=None, save_max=np.inf, save_folder='.',
                 convert=False, click_model_path=None, classifier=None, save_noise=False):
        """
        Process to detect clicks
        Trigger decision
        The trigger automatically makes a measure of background noise and then compares the signal level to the
        noise level. When the signal level reaches a certain threshold above the noise level a click clip is started.
        When the signal level falls to below the threshold for more than a set number of bins, the click clip ends and
        the clip is sent to the localisation and classification modules.
        Parameters
        ----------
        hydrophone : hydrophone object
            Object representing a hydrophone from pyhydrophone
        long_filt : float
            Long filter (used when no click is active)
        long_filt2 : float
            Long filter 2 (used when signal is above threshold)
        short_filt: float
            Short filter parameter
        threshold : float
            Detection threshold in db
        max_length : int
            Maximum length of a click in samples
        threshold : float
            Detection threshold in db
        min_separation : int
            Minimum separation between clicks in samples
        max_length : int
            Maximum length of a click in samples
        pre_samples : int
            Number of samples to store before the click detection
        post_samples : int
            Number of samples to store after the click detection
        fs : int
            Sampling frequency
        prefilter: Filter object
            Prefilter to apply to the signal before passing it to the trigger. It is also applied to the stored clips.
            If set to None a band-pass Butterworth 4th order filter [100000, 150000] Hz will be used
        dfilter : Filter object
            Filter to apply for the trigger. Has to be of the class Filter
            If set to None, a high-pass Butterworth [20000, :] 4th order filter will be used
        save_max : int
            Maximum number of clicks to save in a file
        save_folder : string or Path
            Folder where to save the click output
        classifier : classifier object
            Classifier used to classify the snippets. Needs to have a classify_click function
        save_noise : bool
            Set to True if you want to save the noise clips as well
        """
        self.hydrophone = hydrophone
        # Detector parameters
        self.pre_samples = pre_samples
        self.post_samples = post_samples

        real_min_length = min_length - post_samples - pre_samples
        real_max_length = max_length - post_samples - pre_samples
        if real_min_length < 0:
            real_min_length = 1
            print('This min length is less than pre_samples + post_samples. Setting it to 1...')
        if real_max_length < 0:
            real_max_length = 1000
            print('This max length is less than pre_samples + post_samples. Setting it to 1000...')
        self.triggerfilter = TriggerFilter(long_filt, long_filt2, short_filt, threshold, real_max_length,
                                           real_min_length, min_separation)

        # Initialize the filters. Create them default if None is passed
        if prefilter is None:
            wn = [100000, 150000]
            self.prefilter = Filter(filter_name='butter', order=4, frequencies=wn, filter_type='bandpass', fs=fs)
        else:
            self.prefilter = prefilter
            self.prefilter.fs = fs

        if dfilter is None:
            wn = 20000
            self.dfilter = Filter(filter_name='butter', order=2, frequencies=wn, filter_type='high', fs=fs)
        else:
            self.dfilter = dfilter
            self.dfilter.fs = fs

        self.save_max = save_max
        self.save_folder = save_folder
        self.save_noise = save_noise

        self.columns = ['id', 'datetime', 'filename', 'wave', 'start_sample', 'duration_samples',
                        'duration_us', 'amplitude']
        self.classifier = classifier
        if classifier is not None:
            if not self.check_classifier(classifier):
                raise TypeError('This classifier does not have a function to classify clicks!')
        if convert:
            self.converter = cc.ClickConverter(click_model_path=click_model_path, fs=fs)
            self.columns += self.converter.click_vars
        else:
            self.converter = None

        self.saved_files = []

        # DataFrame with all the clips
        self.clips = pd.DataFrame(columns=self.columns)
        self.clips = self.clips.set_index('id')
        self.columns.remove('id')

        self.fs = fs

        self.split_number = 0

    def __setattr__(self, key, value):
        """
        If the sampling frequency of the sound is different than the one from the filters, update the filters 
        """
        if key == 'fs':
            self.dfilter.fs = value
            self.prefilter.fs = value
            if self.converter is not None:
                self.converter.fs = value
        elif key == 'save_folder':
            value = pathlib.Path(value)
        self.__dict__[key] = value

    @staticmethod
    def check_classifier(classifier):
        if hasattr(classifier, 'classify_click'):
            return True
        return False

    def pre_filter(self, xn):
        """
        Filter the sample xn with the pre-filter

        Parameters
        ----------
        xn : float
            Sample N
        """
        # Apply the prefilter to the sample
        xi = self.prefilter(xn)

        return xi

    def save_clips(self, file_id):
        """
        Save the clips in a file
        """
        clips_filename_h5 = self.save_folder.joinpath('%s_clips.h5' % file_id)
        waves_filename_h5 = self.save_folder.joinpath('%s_waves.h5' % file_id)
        clips_filename_csv = self.save_folder.joinpath('%s_clips.csv' % file_id)

        self.clips.filename = self.clips.filename.astype(str)
        self.clips.datetime = self.clips.datetime.astype(str)
        self.clips.start_sample = self.clips.start_sample.astype(int)
        # Save everything in the file
        if self.classifier is not None:
            if not self.save_noise:
                csv_df = self.clips.drop(index=self.clips.loc[self.clips[self.classifier.class_column] == 3].index)
            else:
                csv_df = self.clips
        else:
            csv_df = self.clips

        csv_df = csv_df.drop(columns=['wave'])
        csv_df.to_csv(clips_filename_csv)

        vlen_type = h5py.special_dtype(vlen=float)
        f = h5py.File(waves_filename_h5, 'w')
        f.create_dataset('/waves', data=self.clips.wave.values, dtype=vlen_type)
        f.close()
        self.clips.drop(columns=['wave']).to_hdf(clips_filename_h5, key='clips')
        self.clips.drop(index=self.clips.index, inplace=True)
        self.saved_files.append(clips_filename_h5)

    def add_click_clips(self, start_sample, date, blocksize, sound_file, clips_list):
        """
        Add all the clips list to the dictionary

        Parameters
        ----------
        start_sample : int
            Sample of the file where the block starts
        date : datetime object
            Datetime where the sound_file starts
        blocksize : int
            Number of frames to read in the block
        sound_file : SoundFile object
            File where the information is stored
        clips_list : list of tuples
            List containing all the clips of clicks as [start_sample, duration] in samples
        """
        sound_file.seek(start_sample)
        signal = sound_file.read(frames=blocksize)

        date += dt.timedelta(seconds=start_sample / sound_file.samplerate)
        # Filter the signal
        filtered_signal = self.dfilter(signal)
        clips_block = self.clicks_block(filtered_signal, date, sound_file.name, start_sample, clips_list)
        self.clips = pd.concat([self.clips, clips_block], ignore_index=True, sort=False)

        # If longer than maximum, save it
        if len(self.clips) >= self.save_max:
            half_file_name = pathlib.Path(sound_file.name).name[:-4] + '_' + str(self.split_number)
            self.save_clips(file_id=half_file_name)
            self.split_number += 1

    def clicks_block(self, signal, date, filename, start_sample_block, clips_list, verbose=False):
        """
        Add the clip to the clip dictionary

        Parameters
        ----------
        signal : np.array
            Signal in the processed block
        date : datetime.datetime
            Datetime where the signal starts
        filename : string
            Name of the file where the signal is
        start_sample_block : int
            First sample of the block according to the whole file
        clips_list: list of tuples
            List containing all the clips of clicks as [start_sample, duration] in samples
        verbose : bool
            Set to True if plots are wanted
        """
        params_matrix = np.zeros((len(clips_list), len(self.columns)))
        timestamps = []
        waves = []
        signal_upa = utils.to_upa(signal, self.hydrophone.sensitivity, self.hydrophone.preamp_gain, self.hydrophone.Vpp)
        for idx, clip in enumerate(clips_list):
            start_sample = clip[0]
            duration_samples = clip[1]
            timestamp = date + dt.timedelta(seconds=start_sample / self.fs)
            # Read the clip
            istart = max(0, start_sample - self.pre_samples)
            frames = min(start_sample - istart + duration_samples + self.post_samples, signal.size)
            clip_upa = signal_upa[istart:istart + frames]
            clip = signal[istart:istart + frames]
            amplitude = utils.amplitude_db(clip_upa)
            timestamps.append(timestamp)
            waves.append(clip_upa)
            if self.converter is not None:
                q, duration, ratio, xc, cf, bw = self.converter.click_params(clip, nfft=512)
                params_matrix[idx, 3:len(self.columns)] = [start_sample_block + istart, frames, frames * 1e6 / self.fs,
                                                           amplitude, q, duration, ratio, xc, cf, bw]
            else:
                params_matrix[idx, 3:len(self.columns)] = [start_sample_block + istart, frames,
                                                           frames * 1e6 / self.fs, amplitude]
            if verbose:
                plt.figure(2, 1)
                plt.plot(clip, label='Filtered signal')
                plt.title('Signal filtered')
                plt.legend()
                plt.tight_layout()
                plt.show()
                plt.close()

        clips_block = pd.DataFrame(params_matrix, columns=self.columns)
        clips_block['wave'] = waves
        clips_block['datetime'] = timestamps
        clips_block['filename'] = filename
        clips_block.start_sample = clips_block.start_sample.astype(np.int32)

        if self.classifier is not None:
            clips_block = self.classifier.classify_matrix(clips_block)
        return clips_block

    def detect_click_clips_file(self, sound_file_path, blocksize=None, date=None, zip_mode=False):
        """
        Return the possible clips containing clicks

        Parameters
        ----------
        sound_file_path : string or Path
            Where the file to be computed is stored
        blocksize : int
            Number of samples to process at a time, if None it will be the length of the file.
            If the blocksize is too small (smaller than 1 period of the lowest frequency of the filter)
            it will affect the results
        date : datetime object
            Date where the file starts. If no date it will be read from the file name.
            Otherwise it will be set up to 01/01/1900 0:0:0
        zip_mode : boolean
            Set to True if the files are zipped
        """
        self.split_number = 0
        # Open file
        if zip_mode:
            folder_zip_file = zipfile.ZipFile(sound_file_path.parent, 'r', allowZip64=True)
            sound_file = sf.SoundFile(folder_zip_file.open(sound_file_path.name))
        else:
            sound_file = sf.SoundFile(sound_file_path, 'r')

        # Update the frequency
        self.fs = sound_file.samplerate

        if blocksize is None:
            blocksize = sound_file.frames

        # Start the filter conditions
        zi = self.prefilter.get_zi()

        # Initialize the clicks count
        n_on = 0
        n_off = 0
        click_on = False

        # Get the initial date of the file
        if date is None:
            try:
                date = self.hydrophone.get_name_datetime(pathlib.Path(sound_file.name).name)
            except ValueError:
                print('Setting date to 01/01/1900 because it was not found')
                date = dt.datetime(1900, 1, 1, 0, 0, 0)

        # Read the file by blocks
        for block_n, block in enumerate(tqdm(sound_file.blocks(blocksize=blocksize, always_2d=True),
                                             total=int(sound_file.frames / blocksize), desc='file', leave=False)):
            prefilter_sig, zi = self.prefilter(block[:, 0], zi=zi)

            # Read samples one by one, apply filter
            clips, click_on, n_on, n_off = self.triggerfilter.update_block(prefilter_sig, click_on, n_on, n_off)
            self.add_click_clips(block_n * blocksize, date, blocksize, sound_file, clips)

        self.save_clips(file_id=sound_file_path.name[:-4])
        return self.clips

    def detect_click_clips_folder(self, folder_path, blocksize=None, zip_mode=False):
        """
        Go through all the sound files and create a database with the detected clicks.

        Parameters
        ----------
        folder_path : string or Path
            Where all the files are
        blocksize : int
            Number of samples to process at a time, if None it will be the length of the file.
            If the blocksize is too small (smaller than 1 period of the lowest frequency of the filter)
            it will affect the results
        zip_mode : bool
            Set to True if the files are zipped
        """
        if isinstance(folder_path, str):
            folder_path = pathlib.Path(folder_path)
        if zip_mode:
            folder_path = zipfile.ZipFile(folder_path, 'r', allowZip64=True)
            files_list = folder_path.namelist()
        else:
            files_list = sorted(folder_path.glob('*.wav'))

        for file_name in tqdm(files_list, total=len(files_list), desc='folder'):
            self.split_number = 0
            # Get the wav
            if zip_mode:
                if file_name.split('.')[-1] == 'wav':
                    wav_file = pathlib.Path(folder_path.filename).joinpath(file_name)
                    try:
                        self.detect_click_clips_file(wav_file, blocksize=blocksize, zip_mode=zip_mode)
                    except RuntimeError as e:
                        print("%s is corrupted and has not been included to the analysis" % wav_file, e)
            else:
                if file_name.suffix == '.wav':
                    wav_file = file_name
                    try:
                        self.detect_click_clips_file(wav_file, blocksize=blocksize, zip_mode=zip_mode)
                    except RuntimeError as e:
                        print("%s is corrupted and has not been included to the analysis" % wav_file, e)

        return self.clips

    def classify_all_saved_clips(self, classifier=None):
        if classifier is not None:
            self.classifier = classifier
        else:
            if self.classifier is None:
                raise Exception('You need to provide one classifier in order to classify!')
        for f in self.saved_files:
            df = pd.read_pickle(f)
            df = self.classifier.classify_matrix(df)
            df.to_hdf5(f, mode='w', key='clips')


spec = [
    ('long_filt', nb.float32),
    ('long_filt2', nb.float32),
    ('short_filt', nb.float32),
    ('threshold', nb.int32),
    ('max_length', nb.int32),
    ('min_length', nb.int32),
    ('min_separation', nb.int32),
    ('Si', nb.float32),
    ('Ni', nb.float32),
    ('trigger', nb.boolean)
]


@nb.experimental.jitclass(spec)
class TriggerFilter:
    def __init__(self, long_filt, long_filt2, short_filt, threshold, max_length, min_length, min_separation):
        """
        Create a Trigger Filter
        The noise level N at sample i is measured using
        Ni = alpha_n * |x|i + (1 - alpha_n) * Ni_1
        and the signal level S is measured using
        Si = alpha_s * |x|i + (1 - alpha_s) * Si_1
        where N is either the long filter parameter when no click is active (i.e. the signal is below threshold)
        or the Long Filter 2 parameter when the signal is above threshold. S is the Short filter parameter.
        A click is started / stopped when the ratio S/N goes above / below the Threshold parameter.

        Parameters
        ----------
        long_filt : float
            Long filter (used when no click is active)
        long_filt2 : float
            Long filter 2 (used when signal is above threshold)
        short_filt: float
            Short filter parameter
        threshold : float
            Detection threshold in db
        min_separation : int
            Minimum separation between clicks in samples
        min_length : int
            Minimum length of a click in samples
        max_length : int
            Maximum length of a click in samples
        """
        self.long_filt = long_filt
        self.long_filt2 = long_filt2
        self.short_filt = short_filt
        self.threshold = 10 ** (threshold / 20.0)
        self.max_length = max_length
        self.min_length = min_length
        self.min_separation = min_separation

        self.Si = 0
        self.Ni = 0

        self.trigger = False

    def initialize_sn(self, block):
        """
        Initialize the values of Si and Ni for correct detection

        Parameters
        ----------
        block : np.array
            Signal to initialize the filter with
        """
        self.Si = np.abs(block).mean() * self.threshold
        self.Ni = np.abs(block).mean()

    def run(self, xi):
        """
        Raise the trigger if SNR is higher than the threshold

        Parameters
        ----------
        xi : float
            Sample i
        """
        # Calculate the SNR 
        self.Si = self.short_filt * np.abs(xi) + (1 - self.short_filt) * self.Si

        if self.trigger:
            self.Ni = self.long_filt2 * np.abs(xi) + (1 - self.long_filt2) * self.Ni
        else:
            self.Ni = self.long_filt * np.abs(xi) + (1 - self.long_filt) * self.Ni

        # Compute SNR = Si/Ni and compare it to the threshold
        if (self.Si / self.Ni) > self.threshold:
            self.trigger = True
        else:
            self.trigger = False

    def update_block(self, prefilter_signal, click_on, n_on, n_off):
        """
        Read all the samples of the block and update the parameters

        Parameters
        ----------
        prefilter_signal : np.array
            Signal to process, already filtered
        click_on : bool
            Set to True if the click was on on the last sample
        n_on : int
            Number of samples the click has been on
        n_off : int
            Number of samples the click has been off
        """
        if click_on:
            start_sample = -n_on
        else:
            start_sample = 0
        clips = []
        i = 0
        for xi in prefilter_signal:
            self.run(xi)
            # If it has been on for too long, save the click!
            if click_on:
                if self.trigger:
                    # Continue the click
                    if n_off > 0:
                        # If it is triggered but the sum of n_on and n_off is already too much
                        # We can't consider it the same click. In case n_on was long enough it is saved
                        # Otherwise it is just discarded
                        if n_on + n_off >= self.max_length:
                            if n_on >= self.min_length:
                                clips.append((start_sample, n_on))
                            click_on = False
                            n_on = 0
                            n_off = 0
                        else:
                            # In case we were already in the count down but it is actually the same click,
                            # consider the count down as part of the click!
                            n_on += n_off
                            n_off = 0
                    n_on += 1
                else:
                    # If it has been off for more than min_separation, save the click! 
                    if n_off >= self.min_separation:
                        if n_on >= self.min_length:
                            clips.append((start_sample, n_on))
                            click_on = False
                            n_on = 0
                            n_off = 0
                    else:
                        # Start to end the click
                        n_off += 1
            else:
                if self.trigger:
                    # Start a new click 
                    start_sample = i
                    click_on = True
                    n_on = 0
                    n_off = 0
            if n_on >= self.max_length:
                clips.append((start_sample, n_on))
                click_on = False
                n_on = 0
                n_off = 0
            i += 1

        return clips, click_on, n_on, n_off


class Filter:
    def __init__(self, filter_name, filter_type, frequencies, order, fs=None):
        """
        Filter object that allows to change the sampling frequency keeping the other parameters
        """
        if filter_name not in ['butter', 'cheby1', 'cheby2', 'besel']:
            filter_name = 'butter'
            raise Warning('Filter %s is not implemented or unknown, setting to Butterworth' % filter_name)
        self.filter_name = filter_name
        self.filter_type = filter_type
        self.frequencies = frequencies
        self.order = order
        self.fs = fs
        if fs is not None:
            self.get_filt(fs)
        else:
            self.filter = None

    def __setattr__(self, key, value):
        if key == 'fs':
            if value is not None:
                self.get_filt(value)
        self.__dict__[key] = value

    def __call__(self, signal, zi=None):
        if zi is None:
            return sig.sosfilt(self.filter, signal)
        else:
            return sig.sosfilt(self.filter, signal, zi=zi)

    def get_filt(self, fs):
        filt = getattr(sig, self.filter_name)
        if type(self.frequencies) is list:
            wn = [f / (fs / 2.0) for f in self.frequencies]
            self.filter = filt(N=self.order, Wn=wn, btype=self.filter_type, analog=False, output='sos')
        else:
            self.filter = filt(N=self.order, Wn=self.frequencies, btype=self.filter_type, analog=False,
                               output='sos', fs=fs)

    def get_zi(self):
        return sig.sosfilt_zi(self.filter)


class ClickDetectorSoundTrapHF(ClickDetector):
    def __init__(self, hydrophone=None, fs=576000, prefilter=None, save_folder='.',
                 convert=False, click_model_path=None, classifier=None, save_noise=False):
        """
        Process to save, filter and classify the clicks from SoundTrap HF detector
        Trigger decision
        The trigger automatically makes a measure of background noise and then compares the signal level to the
        noise level. When the signal level reaches a certain threshold above the noise level a click clip is started.
        When the signal level falls to below the threshold for more than a set number of bins, the click clip ends and
        the clip is sent to the localisation and classification modules.
        Parameters
        ----------
        hydrophone : hydrophone object
            Object representing a hydrophone from pyhydrophone
        fs : int
            Sampling frequency
        prefilter: Filter object
            Prefilter to apply to the signal before passing it to the trigger. It is also applied to the stored clips.
            If set to None a band-pass Butterworth 4th order filter [100000, 150000] Hz will be used
        save_folder : string or Path
            Folder where to save the click output
        classifier : classifier object
            Classifier used to classify the snippets. Needs to have a classify_click function
        save_noise : bool
            Set to True if you want to save the noise clips as well
        """
        if not isinstance(hydrophone, pyhy.SoundTrapHF):
            raise Exception('The hydrophone has to be a SoundTrap with the HF Click detector!')
        super().__init__(hydrophone=hydrophone, fs=fs, prefilter=prefilter, save_folder=save_folder, convert=convert,
                         click_model_path=click_model_path, classifier=classifier, save_noise=save_noise)

    def __setattr__(self, key, value):
        """
        If the sampling frequency of the sound is different than the one from the filters, update the filters
        """
        if key == 'fs':
            self.prefilter.fs = value
            if self.converter is not None:
                self.converter.fs = value
        elif key == 'save_folder':
            value = pathlib.Path(value)
        self.__dict__[key] = value

    def detect_click_clips_file(self, sound_file_path, blocksize=None, date=None, verbose=False, zip_mode=False):
        """
        Return the possible clips containing clicks

        Parameters
        ----------
        sound_file_path : string or Path
            Where the file to be computed is stored
        blocksize : int
            Number of samples to process at a time, if None it will be the length of the file.
            If the blocksize is too small (smaller than 1 period of the lowest frequency of the filter)
            it will affect the results
        date : datetime object
            Date where the file starts. If no date it will be read from the file name.
            Otherwise it will be set up to 01/01/1900 0:0:0
        verbose : boolean
            Set to True to get all the plots from the detections
        zip_mode : boolean
            Set to True if the files are zipped
        """
        self.split_number = 0
        clips = self.hydrophone.read_HFclicks_file(sound_file_path, zip_mode=zip_mode)
        if len(clips) > 0:
            self.fs = clips.iloc[0]['fs']
            params_matrix = np.zeros((len(clips), len(self.columns)))
            print('Calculating parameters and classifying clicks')
            for idx, click in clips.iterrows():
                filtered_wave = self.prefilter(click.wave)
                clip_upa = utils.to_upa(filtered_wave, self.hydrophone.sensitivity, self.hydrophone.preamp_gain,
                                        self.hydrophone.Vpp)
                amplitude = utils.amplitude_db(clip_upa)
                if self.converter is not None:
                    q, duration, ratio, xc, cf, bw = self.converter.click_params(click.wave, nfft=512)
                    params_matrix[idx, 3:len(self.columns)] = [click['start_sample'], click['duration'],
                                                               click['duration'] * 1e6 / self.fs,
                                                               amplitude, q, duration, ratio, xc, cf, bw]
                else:
                    params_matrix[idx, 3:len(self.columns)] = [click['start_sample'], click['duration'],
                                                               click['duration'] * 1e6 / self.fs, amplitude]

                if verbose:
                    fig, ax = plt.subplots(2, 1)
                    ax[0].plot(filtered_wave, label='Filtered signal')
                    ax[1].plot(click.wave, label='Original signal')
                    plt.legend()
                    plt.tight_layout()
                    plt.show()
                    plt.close()
            clips_file = pd.DataFrame(params_matrix, columns=self.columns)
            clips_file['wave'] = clips['wave']
            clips_file['datetime'] = clips.datetime
            clips_file['filename'] = clips.filename
            clips_file.start_sample = clips_file.start_sample.astype(np.int32)
            self.clips = pd.concat([self.clips, clips_file], ignore_index=True, sort=False)
            if self.classifier is not None:
                self.clips = self.classifier.classify_matrix(self.clips)

            self.save_clips(file_id=sound_file_path.name[:-4])
            return self.clips



#!/usr/bin/python
"""
Module : porcc.py
Authors : Clea Parcerisas
Institution : VLIZ (Vlaams Instituut voor de Zee)
Last Accessed : 9/23/2020
"""

__author__ = "Clea Parcerisas"
__version__ = "0.1"
__credits__ = "Clea Parcerisas"
__email__ = "clea.parcerisas@vliz.be"
__status__ = "Development"

import configparser
import itertools
import pickle
from importlib import resources

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn import metrics, linear_model, model_selection

from pyporcc import click_converter, utils

pd.plotting.register_matplotlib_converters()
plt.style.use('ggplot')


class PorCCModel:
    def __init__(self, train_hq_df, train_lq_df, test_df):
        """
        Find the click model
        Parameters
        ----------
        train_hq_df : DataFrame
            DataFrame with the rows used for training the HQ model
        train_lq_df : DataFrame
            DataFrame with the rows used for training the LQ model
        test_df : DataFrame
            DataFrame to test the models
        """
        self.ind_vars = ['Q', 'duration', 'ratio', 'XC', 'CF', 'BW']
        self.dep_vars = ['P']
        self.train_hq_df = train_hq_df
        self.train_lq_df = train_lq_df
        self.test_df = test_df

        # Model definition, initialized to None until calculated 
        self.hq_params = None
        self.hq_mod = None
        self.lq_params = None
        self.lq_mod = None

    def pick_df(self, name):
        """
        Pick the right DataFrame from the model
        Parameters
        ----------
        name : string
            String of the right dataframe to pick. It can be 'hq', 'lq' or 'test'

        Returns
        -------
        DataFrame correspondent to the name
        """
        if name == 'hq':
            df = self.train_hq_df
        elif name == 'lq':
            df = self.train_lq_df
        elif name == 'test':
            df = self.test_df
        else:
            raise Exception('This is not a valid Data Frame name')

        return df

    def hq_lq_separation(self, df, train_size=0.2):
        """
        Separate the df in hq_train, lq_train and test
        Parameters
        ----------
        df : DataFrame
            Whole dataframe to be separated
        train_size : float
            From 0 to 1, percentage to use to train

        Returns
        -------
        DataFrame for hq model, DataFrame for loq model and DataFrame for testing
        """
        train, self.test_df = model_selection.train_test_split(df, train_size=train_size)
        self.train_hq_df = train[train['P'] != 2]
        self.train_lq_df = train[train['P'] != 1]
        self.train_hq_df['P'].replace(3, 0)
        self.train_lq_df['P'].replace(3, 0)
        self.train_lq_df['P'].replace(2, 1)

        return self.train_hq_df, self.train_lq_df, self.test_df

    def load_model_from_config(self, configfile_path):
        """
        Load PorCC model coefficients
        Parameters
        ----------
        configfile_path : string or Path
            .ini file where the coefficients of the LQ and the HQ models are specified
        """
        config = configparser.ConfigParser().read(configfile_path)

        logitcoef_hq = np.array(config['MODEL']['logitCoefHQ'].split(',')).astype(float)
        logitcoef_lq = np.array(config['MODEL']['logitCoefLQ'].split(',')).astype(float)

        self.hq_params = np.array(config['MODEL']['hq_params'].split(','))
        self.lq_params = np.array(config['MODEL']['lq_params'].split(','))

        # Starts and fit to get the classes 
        logit_hq = linear_model.LogisticRegression()
        reg_hq = logit_hq.fit(self.train_hq_df[self.hq_params], self.train_hq_df['P'])

        logit_lq = linear_model.LogisticRegression()
        reg_lq = logit_lq.fit(self.train_lq_df[self.lq_params], self.train_lq_df['P'])

        # Cheat and force the coefficients
        reg_hq.coef_ = np.array([logitcoef_hq[:-1]])
        reg_hq.intercept_ = logitcoef_hq[-1]
        self.hq_mod = reg_hq

        reg_lq.coef_ = np.array([logitcoef_lq[:-1]])
        reg_lq.intercept_ = logitcoef_lq[-1]
        self.lq_mod = reg_lq

    def find_best_model(self, name):
        """
        Find the best model among the possible models
        Parameters
        ----------
        name : string
            Set to 'HQ' or 'LQ' to find the according model

        Returns
        -------
        columns, model. Columns are the names of the relevant parameters of the model. Model is an object instance
        """
        # Get all the possible models
        models = self.find_possible_models(name)
        df = self.pick_df(name)

        # Select the appropriate columns combination according to AIC
        models = models[models[:, -1].argsort()]
        comb, mod, aic = models[0]

        # y = df['P']
        x = df[comb]

        if name == 'hq':
            self.hq_mod = mod
            self.hq_params = x.columns
        else:
            self.lq_mod = mod
            self.lq_params = x.columns

        print('The winning combination for %s is %s. AIC: %s' % (name, comb, aic))
        # print(mod.summary())
        return x.columns, mod

    def find_possible_models(self, name):
        """
        Create all the regression models
        classification ('P') = 0:N, 1:LQ, 2:HQ

        Parameters
        ----------
        name : string
            Set to 'HQ' or 'LQ' to find the according model

        Returns
        -------
        List of all the possible models as instances (alreay fitted)
        """
        models = []
        df = self.pick_df(name)

        # Go through all the possible combinations (from 1 to all the variables)
        for i in np.arange(1, len(self.ind_vars) + 1):
            var_combinations = itertools.combinations(self.ind_vars, i)
            for comb in var_combinations:
                # Regression model
                y = df['P']
                x = df[list(comb)]

                logit = linear_model.LogisticRegression(max_iter=500, tol=1e-5, C=0.1)
                reg = logit.fit(x, y)

                # Calculate AIC
                y_prob = reg.predict_proba(x)
                aic = utils.aic_score(y, y_prob, len(comb))

                # Append the model
                models.append([list(comb), reg, aic])

        return np.array(models)

    def save(self, save_path):
        """
        Save the current models in a file. It will saved as a pickle

        Parameters
        ----------
        save_path : string
            Path where to save the models
        """
        pickle.dump(self, open(save_path, 'wb'))

    def calculate_clicks_params(self, df_name, fs, click_model_path, save_path=None):
        """
        Add to the df the click parameters calculated by the Click Class

        Parameters
        ----------
        df_name : string
            name of the dataframe to pick. It can be 'hq', 'lq' or 'test'
        fs : int
            Sampling frequency of the dataframe entries
        click_model_path : string or Path
            Path to the wav file containing the click model
        save_path : string or Path
            Path where to save the output (should be pickle extension)
        """
        # Pick the right df
        df = self.pick_df(df_name)

        # Pass the sampling frequency as metadata
        df.fs = fs

        # Init a converter to calculate all the params
        converter = click_converter.ClickConverter(click_model_path, self.ind_vars)
        df_clicks = converter.clicks_df(df)

        df_joint = df.join(df_clicks, lsuffix='_mat', rsuffix='')
        if save_path is not None:
            df_joint.to_pickle(save_path)

        return df_joint


class PorCC:
    def __init__(self, load_type, class_column='pyPorCC', **kwargs):
        """
        Start the classifier
        If load_type is set to 'manual', loads the models from the config file. Then config_file must be specified
        If load_type is set to 'trained_model', loads the trained model.
        Then hq_mod, lq_mod, hq_params, lq_params have to be specified

        load_type : string
            'manual' or 'trained_model'
        """
        self.th1 = 0.9999  # threshold for HQ clicks
        self.th2 = 0.55  # threshold for LQ clicks
        self.lowcutfreq = 100e3  # Lowcut frequency
        self.highcutfreq = 160e3  # Highcut frequency
        self.class_column = class_column

        self.load_type = load_type
        if load_type == 'manual':
            if kwargs['config_file'] == 'default':
                with resources.path('pyporcc.models', 'log_models.ini') as config_path:
                    self.manual_models(config_path)
            else:
                self.manual_models(kwargs['config_file'])
        else:
            for key, val in kwargs.items():
                self.__dict__[key] = val

    def manual_models(self, configfile_path):
        """
        Load the coefficients of the models to calculate the probability manually

        Parameters
        ----------
        configfile_path : string or Path
            Path to the config file with the coefficients for the models
        """
        config = configparser.ConfigParser()
        config.read(configfile_path)

        hq_coef = np.array(config['MODEL']['logitCoefHQ'].split(',')).astype(float)
        lq_coef = np.array(config['MODEL']['logitCoefLQ'].split(',')).astype(float)

        self.hq_mod = ManualLogit(hq_coef)
        self.lq_mod = ManualLogit(lq_coef)

        self.hq_params = np.array(config['MODEL']['hq_params'].split(','))
        self.lq_params = np.array(config['MODEL']['lq_params'].split(','))

    def classify_click(self, click):
        """
        Classify the click in HQ, LQ, N

        Parameters
        __________
        click : Click object
            Click to classify
        """
        x = pd.DataFrame(data={'Q': click.Q, 'duration': click.duration, 'ratio': click.ratio, 'XC': click.xc,
                               'CF': click.cf, 'PF': click.pf, 'BW': click.bw})
        porps = self.classify_row(x)

        return porps

    def classify_row(self, x):
        """
        Classify one row according to the params [PF, CF, Q, XC, duration, ratio, BW]

        Parameters
        ----------
        x : pandas row or dictionary
            Row to be classified
        """
        # Add the independent variable
        x.at['const'] = 1
        if (x['CF'] > self.lowcutfreq) and (x['CF'] < self.highcutfreq) and (x['Q'] > 4):
            # Evaluate the model on the given x
            prob_hq = self.hq_mod.predict_proba(x[self.hq_params])[0][1]

            # Assign clip to a category
            if prob_hq >= self.th1:
                # HQ click
                porps = 1
            else:
                prob_lq = self.lq_mod.predict_proba(x[self.lq_params])[0][1]
                if prob_lq > self.th2:
                    # LQ click
                    porps = 2
                else:
                    # HF Noise
                    porps = 3
        else:
            porps = 3

        return porps

    def classify_matrix(self, df):
        """
        Classify according to the params [PF, CF, Q, XC, duration, ratio, BW]

        Parameters
        ----------
        df : DataFrame
            DataFrame to be classified. Parameters PF, CF, Q, XC, duration, ratio and BW must be specified
        """
        # Add the independent variable for the regression
        # Initialize the prediction column
        df = df.assign(const=1)
        df[self.class_column] = 0

        # Evaluate the model on the given x
        df = df.assign(prob_hq=self.hq_mod.predict_proba(df[self.hq_params])[:, 1],
                       prob_lq=self.lq_mod.predict_proba(df[self.lq_params])[:, 1])

        # Decide
        loc_idx = (df['CF'] > self.lowcutfreq) & (df['CF'] < self.highcutfreq) & (df['Q'] > 4)
        # Add remove duration > 450
        df.loc[~loc_idx, self.class_column] = 3  # N Clicks
        df.loc[loc_idx & (df['prob_hq'] > self.th1), self.class_column] = 1  # HQ Clicks
        df.loc[loc_idx & (df['prob_hq'] < self.th1) & (df['prob_lq'] > self.th2), self.class_column] = 2  # LQ Clicks
        df.loc[loc_idx & (df['prob_hq'] < self.th1) & (df['prob_lq'] <= self.th2), self.class_column] = 3  # N Clicks

        return df.drop(columns=['const'])

    def predict(self, df):
        """
        Classify and return the y value

        Parameters
        ----------
        df : DataFrame

        """
        y_pred = self.classify_matrix(df)[self.class_column]

        return y_pred

    def predict_proba(self, df):
        """
        Return the probability of being classified as HQ or LQ

        Parameters
        ----------
        df : DataFrame
            DataFrame to predict. Needs to have the columns passed as hq and lq parameters

        Returns
        -------
        DataFrame with prob_hq, prob_lq and total PorCC probability
        """
        # Add the independent variable for the regression
        df = df.assign(const=1)

        # Initialize the prediction column
        df = df.assign(prob=0)

        # Evaluate the model on the given x
        df = df.assign(prob_hq=self.hq_mod.predict_proba(df[self.hq_params])[:, 1])
        df = df.assign(prob_lq=self.lq_mod.predict_proba(df[self.lq_params])[:, 1])

        # Decide
        loc_idx = (df['CF'] > self.lowcutfreq) & (df['CF'] < self.highcutfreq) & (df['Q'] > 4)
        df.loc[~loc_idx, 'prob'] = 0
        df.loc[loc_idx & (df['prob_hq'] > self.th1), 'prob'] = df.loc[loc_idx & (df['prob_hq'] > self.th1), 'prob_hq']
        df.loc[loc_idx & (df['prob_hq'] < self.th1), 'prob'] = df.loc[loc_idx & (df['prob_hq'] < self.th1), 'prob_lq']

        return df[['prob_hq', 'prob_lq', 'prob']]

    def test_classification(self, test_df, col_name='ClassifiedAs'):
        """
        Test the algorithm. With the same parameters, test the prediction output of the algorithm

        Parameters
        ----------
        test_df : DataFrame
            DataFrame to test, with the necessary columns for classify
        col_name : string
            Column name of the correct classification
        """
        predicted_df = self.classify_matrix(test_df)

        # Compare self.class_column vs 'ClassifiedAs' (MATLAB)
        error = np.sum(test_df[col_name] != predicted_df[self.class_column]) / len(test_df)

        return error, predicted_df

    def plot_roc_curves(self, df, dep_var='ManualAsign'):
        """
        Plot the ROC curves for HQ, LQ and All

        Parameters
        ---------
        df : DataFrame
            DataFrame with the classification
        dep_var : string
            Name of the column where the classification
        """
        fig, ax = plt.subplots(1, 3)
        # Filter the data frame so it only has HQ-Noise, LQ-Noise and All-Noise
        hq_noise = df.loc[df[dep_var] != 2]
        lq_noise = df.loc[df[dep_var] != 1]
        self._plot_roc_curves(df=hq_noise, dep_var=dep_var, ax0=ax[0])
        ax[0].set_title('HQ vs Noise')
        self._plot_roc_curves(df=lq_noise, dep_var=dep_var, ax0=ax[1])
        ax[1].set_title('LQ vs Noise')
        self._plot_roc_curves(df=df, dep_var=dep_var, ax0=ax[2])
        ax[2].set_title('All vs Noise')

        plt.tight_layout()
        plt.show()
        plt.close()

    def _plot_roc_curves(self, df, dep_var='ManualAsign', ax0=None):
        """`    
        Plot the ROC curves for each part of the algorithm

        Parameters
        ----------
        df : DataFrame
            DataFrame with the information
        dep_var : string
            Name of the column of the dependent variable where the correct labels are
        """
        y_test = self.convert2binary(df[dep_var])
        if ax0 is None:
            fig, ax = plt.subplots()
        else:
            ax = ax0
        probs = self.predict_proba(df)

        # Plot ROC
        fpr_hq, tpr_hq, thresholds = metrics.roc_curve(y_test, probs['prob_hq'], drop_intermediate=True)
        fpr_lq, tpr_lq, thresholds = metrics.roc_curve(y_test, probs['prob_lq'], drop_intermediate=True)
        fpr_all, tpr_all, thresholds = metrics.roc_curve(y_test, probs['prob'], drop_intermediate=True)

        ax.plot(fpr_hq, tpr_hq, label='PorCC HQ')
        ax.plot(fpr_lq, tpr_lq, label='PorCC LQ')
        ax.plot(fpr_all, tpr_all, label='PorCC All')

        if ax0 is None:
            plt.tight_layout()
            plt.show()
            plt.close()

        return ax

    @staticmethod
    def convert2binary(y):
        """
        Return the y sample as binary (no porpoise / porpoise)

        Parameters
        ----------
        y : np.array
            Array with 3 classes to be converted to 2 classes (from Noise, LQ and HQ to Porpoise/No Porpoise)
        """
        y = y.replace(3, 0)
        y = y.replace(2, 1)

        return y


class ManualLogit:
    def __init__(self, coef, th=0.5):
        """
        Init a logit probability prediction class

        Parameters
        ----------
        coef : list or np.array
            Coefficients of the model
        th : float
            From 0 to 1 threshold to make the classification decision
        """
        self.coef_ = coef
        self.th_ = th

    def predict(self, x):
        """
        Predict the classification of X

        Parameters
        ----------
        x : np.array
            Array with the X coefficients to be predict the class
        """
        proba = self.predict_proba(x)[0][:, 1]
        y_pred = np.zeros(proba.shape)
        y_pred[np.where(proba >= self.th_)] = 1
        return y_pred

    def predict_proba(self, x):
        """
        Predict the probability with the coefficients

        Parameters
        ----------
        x : np.array
            Array with the X coefficients to be predict the probability
        """
        lower_bnd = np.log(np.finfo(float).eps)
        upper_bnd = -lower_bnd
        if len(x.shape) == 1:
            xb = (x * self.coef_[1:]).sum() + self.coef_[0]
        else:
            xb = (x * self.coef_[1:]).sum(axis=1) + self.coef_[0]
        odds = np.exp(-utils.constrain(xb, lower_bnd, upper_bnd))
        prob_1 = 1 / (1 + odds)
        prob = np.column_stack((1 - prob_1, prob_1))
        return prob

    def predict_log_proba(self, x):
        """
        Predict the log probability

        Parameters
        ----------
        x : np.array
            Array with the X coefficients to be predict the logarithmic probability
        """
        prob = self.predict_proba(x)
        log_prob = np.log(prob)

        return log_prob
